﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Windows.Forms;

namespace LonScan
{
    public partial class Form1 : Form
    {
        private int Address = 10;
        private Thread ScanThread = null;
        private Dictionary<int, ResponseCallback> PendingRequests = new Dictionary<int, ResponseCallback>();

        public delegate void ResponseCallback(byte[] response);
        private int NextTransaction = 0;

        private int MemoryAddress = 0;

        public string[][] TypeDescPMX =
        {
            new string[] { "nviRequest       ", "                 ", "SNVT_obj_request" },
            new string[] { "nvoStatus        ", "                 ", "SNVT_obj_status" },
            new string[] { "nviTimeSet       ", "nvi_Systemzeit   ", "SNVT_time_stamp" },
            new string[] { "nvoFileDirectory ", "                 ", "SNVT_address" },
            new string[] { "nvoTime          ", "Lokale Zeit      ", "SNVT_time_stamp" },
            new string[] { "nviObjEvent      ", "                 ", "" },
            new string[] { "nvoObjStatus     ", "                 ", "" },
            new string[] { "nvoError         ", "                 ", "" },
            new string[] { "nvoWvEnergyHold  ", "nvo_EnergyHold_HK", "SNVT_lev_percent" },
            new string[] { "nvoBwEnergyHold  ", "nvo_EnergyHold_BK", "SNVT_lev_percent" },
            new string[] { "nviWE_ModeGp1    ", "nvi_WE_Mode_Gr1  ", "" },
            new string[] { "nviWE_ModeGp3    ", "nvi_WE_Mode_Gr3  ", "" },
            new string[] { "EH_m             ", "                 ", "" },
            new string[] { "WET_nviTsoll[0]  ", "WET_T_Soll_HK    ", "SNVT_temp_p" },
            new string[] { "WET_nviTsoll[1]  ", "WET_T_Soll_BK    ", "SNVT_temp_p" },
            new string[] { "WET_nvoTist      ", "nvo_WET_T_Kessel ", "SNVT_temp_p" },
            new string[] { "BD_nviDsp        ", "                 ", "display" },
            new string[] { "GB_nviNsoll      ", "                 ", "SNVT_rpm" },
            new string[] { "GB_nvoNist       ", "A_GB_Drehzahl    ", "SNVT_rpm" },
            new string[] { "GB_nvoNsoll      ", "A_GB_Drehz. Soll ", "SNVT_rpm" },
            new string[] { "GB_m             ", "                 ", "" },
            new string[] { "FS_nviMsoll      ", "A_FS_Menge_Soll  ", "SNVT_mass_kilo" },
            new string[] { "FS_nvoMotor      ", "                 ", "" },
            new string[] { "FS_nvoTein       ", "                 ", "" },
            new string[] { "FS_nvoTaus       ", "                 ", "" },
            new string[] { "FS_nvoMpts       ", "                 ", "" },
            new string[] { "FS_m             ", "                 ", "" },
            new string[] { "FS_nviMfoerder   ", "A_FS_Menge_berech", "SNVT_mass_kilo" },
            new string[] { "FS_nvoAvgTb_Tk   ", "                 ", "" },
            new string[] { "ZG_nviOn         ", "                 ", "" },
            new string[] { "ZG_nvoHeat       ", "                 ", "" },
            new string[] { "ZG_nvoVent       ", "                 ", "" },
            new string[] { "RO_nviCount      ", "                 ", "" },
            new string[] { "RO_nvoCount      ", "                 ", "" },
            new string[] { "RO_m             ", "                 ", "" },
            new string[] { "PMX_nviLstg      ", "P_Leistung_Soll  ", "SNVT_lev_cont" },
            new string[] { "FMF_nvoStatus    ", "                 ", "" },
            new string[] { "PMX_nvoLstg      ", "P_Leistung       ", "SNVT_lev_cont" },
            new string[] { "PMX_m            ", "                 ", "" },
            new string[] { "PMX_nvoToPzs     ", "                 ", "" },
            new string[] { "PMX_nviFromPzs   ", "                 ", "" },
            new string[] { "PMX_nviPzsStatus ", "                 ", "" },
            new string[] { "PMX_eeBetrStd    ", "B_Betriebsstunden", "SNVT_time_hour" },
            new string[] { "PMX_eeNbrAnhz    ", "B_Anheizvorgänge ", "SNVT_count" },
            new string[] { "EXT_m            ", "B_AAT_HFR        ", "" },
            new string[] { "EXT_nvoTacho     ", "                 ", "" },
            new string[] { "EXT_nvoTacho1    ", "                 ", "" },
            new string[] { "NIC_nvoValue     ", "T_Brennraum      ", "SNVT_temp" },
            new string[] { "NIC_nvoTboard    ", "T_Schaltfeld     ", "SNVT_temp_p" },
            new string[] { "NIC_nvoAvgVal    ", "T_Brennr._Mittel ", "SNVT_temp" },
            new string[] { "TK_nviSetP       ", "T_Kessel_Soll    ", "SNVT_temp_p" },
            new string[] { "TK_nvoPID        ", "                 ", "" },
            new string[] { "TK_nvoTemp       ", "T_Kessel         ", "SNVT_temp_p" },
            new string[] { "TK_nvoRist       ", "T_Kessel_Rücklauf", "SNVT_temp_p" },
            new string[] { "RG_nviSetP       ", "T_Abgas_Soll     ", "SNVT_temp" },
            new string[] { "RG_nvoPID        ", "                 ", "" },
            new string[] { "RG_nvoTemp       ", "T_Abgas          ", "SNVT_temp" },
            new string[] { "TVB_nvoValue     ", "                 ", "" },
            new string[] { "HW_nviDigInput   ", "                 ", "" },
            new string[] { "HW_nviDigOutput  ", "                 ", "" }
        };


        public string[][] TypeDescUML =
        {
            new string[] { "nviRequest         ", "                 ", "SNVT_obj_request" },
            new string[] { "nvoStatus          ", "                 ", "SNVT_obj_status" },
            new string[] { "nviTimeSet         ", "nvi_Systemzeit   ", "SNVT_time_stamp" },
            new string[] { "nvoFileDirectory   ", "                 ", "SNVT_address" },
            new string[] { "nvoTimeSet         ", "Lokale Zeit      ", "SNVT_time_stamp" },
            new string[] { "nviError           ", "nvi_Störcode     ", "SNVT_count" },
            new string[] { "nvoError           ", "nvo_Störcode     ", "SNVT_count" },
            new string[] { "nciLocation        ", "Bezeichnung      ", "SNVT_str_asc" },
            new string[] { "N35_m              ", "                 ", "UNVT" },
            new string[] { "N35_nviDigOut      ", "                 ", "SNVT_state" },
            new string[] { "N35_nvoDiRaw       ", "                 ", "UNVT" },
            new string[] { "N35_nvoAdRaw       ", "                 ", "UNVT" },
            new string[] { "N35_nvoBws         ", "                 ", "UNVT" },
            new string[] { "nviWvEnergyHold    ", "nvi_EnergyHold_HK", "SNVT_lev_percent" },
            new string[] { "nviBwEnergyHold    ", "nvi_EnergyHold_BK", "SNVT_lev_percent" },
            new string[] { "nviWE_ModeGp1      ", "nvi_WE_Mode_Gr1  ", "SNVT_hvac_mode" },
            new string[] { "EH_m               ", "                 ", "UNVT" },
            new string[] { "WET_nviTsoll[0]    ", "WET_T_Soll_HK    ", "SNVT_temp_p" },
            new string[] { "WET_nviTsoll[1]    ", "WET_T_Soll_WW    ", "SNVT_temp_p" },
            new string[] { "WET_nvoTsoll[0]    ", "WET_T_Soll_HK    ", "SNVT_temp_p" },
            new string[] { "WET_nvoTsoll[1]    ", "WET_T_Soll_WW    ", "SNVT_temp_p" },
            new string[] { "WET_nviTist        ", "nvi_WE_T_Kessel  ", "SNVT_temp_p" },
            new string[] { "WET_nvoTist        ", "nvo_WET_T_Kessel ", "SNVT_temp_p" },
            new string[] { "WET_m              ", "                 ", "UNVT" },
            new string[] { "nviTaFb            ", "nvi_T_Aussen     ", "SNVT_temp_p" },
            new string[] { "nvoTa              ", "T_Aussen         ", "SNVT_temp_p" },
            new string[] { "TA_m               ", "                 ", "UNVT" },
            new string[] { "LX_nviTist[0]      ", "L_T_Boiler       ", "SNVT_temp_p" },
            new string[] { "LX_nviTsoll[0]     ", "L_T_Boiler1_Soll ", "SNVT_temp_p" },
            new string[] { "LX_nvoPump[0]      ", "L_Pumpe_BK1      ", "SNVT_lev_cont" },
            new string[] { "LX_nvoValve[0]     ", "L_Ladeventil_BK1 ", "SNVT_lev_cont" },
            new string[] { "LX_m[0]            ", "                 ", "UNVT" },
            new string[] { "LX_nvoSetPt        ", "                 ", "SNVT_temp_p" },
            new string[] { "WVF_nviTPO         ", "T_PufferOben TPO ", "SNVT_temp_p" },
            new string[] { "M_nviTVsoll        ", "T_Vorlauf_Soll   ", "SNVT_temp_p" },
            new string[] { "M_nviTVist         ", "T_Vorlauf        ", "SNVT_temp_p" },
            new string[] { "M_nvoPump          ", "Pumpe            ", "SNVT_lev_cont" },
            new string[] { "M_nvoValve         ", "M_Mischventil    ", "SNVT_lev_percent" },
            new string[] { "M_m                ", "                 ", "UNVT" },
            new string[] { "RC_nviMode[0]      ", "                 ", "UNVT" },
            new string[] { "RC_nvoMode[0]      ", "BDM1_Betriebsart ", "UNVT" },
            new string[] { "RC_nviFlags[0]     ", "                 ", "UNVT" },
            new string[] { "RC_nvoFlags[0]     ", "                 ", "UNVT" },
            new string[] { "RC_nviError[0]     ", "BDM_Störcode     ", "SNVT_count" },
            new string[] { "RC_nvoSpaceTemp[0] ", "BDM_T_Raum       ", "SNVT_temp_p" },
            new string[] { "RC_nvoHeatSetPt[0] ", "                 ", "SNVT_temp_p" },
            new string[] { "RC_nvoBoilSetPt[0] ", "BDM_T_Boiler_Soll", "SNVT_temp_p" },
            new string[] { "RC_nviSolarTemp    ", "BDM_TSA_Koll.Aust", "SNVT_temp_p" },
            new string[] { "RC_nviSolarEnrgy   ", "BDM_Solargewinn  ", "SNVT_elec_kwh" },
            new string[] { "RC_nviSoDayEnrgy   ", "                 ", "SNVT_elec_kwh" },
            new string[] { "FA_m               ", "                 ", "UNVT" },
            new string[] { "FA_nviTVsoll       ", "FA_T_Kessel_Soll ", "SNVT_temp_p" },
            new string[] { "FA_nvoTk           ", "FA_T_Kessel_VL   ", "SNVT_temp_p" },
            new string[] { "FA_nvoTr           ", "FA_T_Kessel_RL   ", "SNVT_temp_p" },
            new string[] { "FA_nviExtTsoll     ", "                 ", "SNVT_temp_p" },
            new string[] { "SYS_pack           ", "                 ", "UNVT" },
        };


        public Form1()
        {
            InitializeComponent();

            listView1.ContextMenu = new ContextMenu(new MenuItem[] { new MenuItem("Set value", (e, s) =>
            {
                if(listView1.SelectedItems.Count == 0)
                {
                    return;
                }

                var item = listView1.SelectedItems[0];

                if(item.Tag == null)
                {
                    return;
                }
                int nv = (int)item.Tag;

                ScanMemory(nv);
                /*
                var form = new HexForm(item.SubItems[3].Text);

                if(form.ShowDialog() == DialogResult.OK)
                {
                }
                */
            }
            ) });
        }

        private void ScanMemory(int nv)
        {
            Thread t = new Thread(()=>
            {
                MemoryAddress = 0x2306;

                while (MemoryAddress < 0x2400)
                {
                    byte[] sendbuf = { 0x01, 0x19, 0x01, 0xFE, 0x01, 0x00, 0x54, 0x00, 0x60, 0x00, 0x00, 0x00, 0x00 };

                    sendbuf[5] = (byte)(Address | 0x80);
                    sendbuf[8] |= (byte)(13); /* read memory */
                    sendbuf[9] = (byte)0x25;
                    sendbuf[10] = (byte)(MemoryAddress >> 8);
                    sendbuf[11] = (byte)(MemoryAddress & 0xFF);
                    sendbuf[12] = (byte)(16);

                    bool success = SendMessage(sendbuf, (d) =>
                    {
                        Console.WriteLine("M: " + Address.ToString("X4") + "  " + BitConverter.ToString(d).Replace("-", " "));
                    });

                    if (success)
                    {
                        MemoryAddress += 0x10;
                    }
                }
            });
            ScanThread.Abort();
            t.Start();
        }

        private bool SendMessage(byte[] data, ResponseCallback response)
        {
            Socket sock = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            IPAddress broadcast = IPAddress.Parse("192.168.1.255");
            IPEndPoint ep = new IPEndPoint(broadcast, 3334);

            int trans = NextTransaction;

            NextTransaction++;
            NextTransaction %= 8;

            lock (PendingRequests)
            {
                PendingRequests.Add(trans, response);
            }

            data[7] &= 0xF0;
            data[7] |= (byte)trans;

            sock.SendTo(data, ep);

            DateTime start = DateTime.Now;

            lock (PendingRequests)
            {
                while ((DateTime.Now - start).TotalMilliseconds < 2000)
                {
                    Monitor.Wait(PendingRequests, 50);

                    if(!PendingRequests.ContainsKey(trans))
                    {
                        return true;
                    }
                }
                PendingRequests.Remove(trans);
            }

            return false;
        }

        protected override void OnClosing(CancelEventArgs e)
        {
            if(ScanThread != null)
            {
                ScanThread.Abort();
                ScanThread = null;
            }
            base.OnClosing(e);
        }

        protected override void OnLoad(EventArgs e)
        {
            base.OnLoad(e);

            Thread logger = new Thread(ThreadReceive);
            logger.Start();

            //SetTime(90);
        }

        private void SetTime(int v)
        {
            byte[] sendbuf = { 0x01, 0x19, 0x01, 0xFE, 0x01, 0x00, 0x54, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 };

            sendbuf[5] = (byte)(Address | 0x80);
            sendbuf[8] |= (byte)(14); /* write parameter */
            sendbuf[9] = (byte)0x25;
            sendbuf[10] = 0x06;
            sendbuf[11] = 0x23;
            sendbuf[12] = 0x06;
            sendbuf[13] = 0x02;
            sendbuf[14] = 0xC8;
            sendbuf[15] = 0x00;
            sendbuf[16] = 0x40;
            sendbuf[17] = (byte)(v & 0xFF);
            sendbuf[18] = (byte)(v >> 8);

            bool success = SendMessage(sendbuf, (d) =>
            {
                Console.WriteLine("M: " + Address.ToString("X4") + "  " + BitConverter.ToString(d).Replace("-", " "));
            });

        }

        void ThreadMain()
        {
            this.BeginInvoke(new Action(() => { textBox1.Text = ""; }));
            while (true)
            {
                for (int nv = 0x0; nv < 0x40; nv++)
                //int nv = 1;
                {
                    byte[] sendbuf = { 0x01, 0x19, 0x01, 0xFE, 0x01, 0x8b, 0x54, 0x00, 0x73, 0x35 };

                    sendbuf[5] = (byte)(Address | 0x80);
                    sendbuf[9] = (byte)nv;

                    bool success = SendMessage(sendbuf, (d) =>
                    {
                    });

                    Thread.Sleep(75);
                }
                Thread.Sleep(250);
            }
        }

        void ThreadReceive()
        {
            UdpClient listener = new UdpClient(3333);
            IPEndPoint groupEP = new IPEndPoint(IPAddress.Any, 3333);

            listener.Client.ReceiveBufferSize = 1024 * 1024;

            while (true)
            {
                byte[] bytes = listener.Receive(ref groupEP);

                if(bytes[0] != 2)
                {
                    continue;
                }

                if (bytes.Length >= 16)
                {
                    int trans = bytes[6 + 7] & 0x0F;
                    bool response = (bytes[6 + 7] & 0x20) != 0;

                    lock (PendingRequests)
                    {
                        if (response && PendingRequests.ContainsKey(trans))
                        {
                            PendingRequests[trans](bytes);
                            PendingRequests.Remove(trans);
                            Monitor.Pulse(PendingRequests);
                        }
                    }

                    if ((bytes[13] & 0xF0) == 0x20 && (bytes[14] == 0x33) && bytes[9] == (byte)(Address | 0x80))
                    {
                        int nv_recv = bytes[15];
                        string name = "";
                        string desc = "";
                        string hex = "";
                        string value = "";
                        string type = "";
                        int len = bytes.Length - 2 - 16;

                        for (int pos = 16; pos < bytes.Length - 2; pos++)
                        {
                            hex += bytes[pos].ToString("X2");
                            if ((pos % 2) == 1)
                            {
                                hex += " ";
                            }
                        }

                        if (len == 2)
                        {
                            int tempRaw = bytes[16] << 8 | bytes[16 + 1];
                            float temp = (float)tempRaw / 100;

                            value = "Temp(?): " + temp + " °C";


                            float temp2 = (float)tempRaw / 10 - 273.15f;

                            value += " Temp(?): " + temp2 + " °C";
                        }

                        /* UML C1 modules */
                        if (Address == 10 || Address == 11)
                        {
                            name = TypeDescUML[nv_recv][0];
                            desc = TypeDescUML[nv_recv][1];
                            type = TypeDescUML[nv_recv][2];
                        }

                        /* PMX */
                        if (Address == 60)
                        {
                            name = TypeDescPMX[nv_recv][0];
                            desc = TypeDescPMX[nv_recv][1];
                            type = TypeDescPMX[nv_recv][2];
                        }


                        switch (type)
                        {
                            case "display":
                                {
                                    value = "" + (char)bytes[16] + (char)bytes[16 + 1] + " " + bytes[16 + 2].ToString();
                                    break;
                                }
                            case "SNVT_count":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];

                                    value = "" + raw;
                                    break;
                                }
                            case "SNVT_power":
                                {
                                    int raw = bytes[16];

                                    value = "Power: " + raw;
                                    break;
                                }
                            case "SNVT_rpm":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];

                                    value = "RPM: " + raw;
                                    break;
                                }
                            case "SNVT_time_hour":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];

                                    value = "" + raw + "h";
                                    break;
                                }
                            case "SNVT_mass_kilo":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];

                                    value = "Kilogramm: " + raw;
                                    break;
                                }
                            case "SNVT_temp":
                                {
                                    int tempRaw = bytes[16] << 8 | bytes[16 + 1];

                                    if (tempRaw < 0x7FFF)
                                    {
                                        float temp = (float)tempRaw / 10 - 273.15f;

                                        value = temp + " °C";
                                    }
                                    else
                                    {
                                        value = "(N/A)";
                                    }
                                    break;
                                }
                            case "SNVT_temp_p":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];
                                    short tempRaw = (short)raw;

                                    if (tempRaw < 0x7FFF)
                                    {
                                        float temp = (float)tempRaw / 100;

                                        value = temp + " °C";
                                    }
                                    else
                                    {
                                        value = "(N/A)";
                                    }
                                    break;
                                }
                            case "SNVT_lev_percent":
                                {
                                    int raw = bytes[16] << 8 | bytes[16 + 1];
                                    float temp = (float)raw / 200;

                                    value = "pct: " + temp.ToString("0.00") + " %";
                                    break;
                                }
                            case "SNVT_lev_cont":
                                {
                                    int raw = bytes[16];
                                    float temp = (float)raw / 2.0f;

                                    value = "lvl: " + temp + " %";
                                    break;
                                }
                            case "":
                                {
                                    value = "";
                                    break;
                                }
                            default:
                                {
                                    value = "unknown: " + type;
                                    break;
                                }
                        }
                        SubmitValue(nv_recv, name, desc, hex, value);
                    }
                }
            }

            listener.Close();
        }

        private void SubmitValue(int nv_recv, string name, string desc, string hex, string value)
        {
            try
            {

                this.BeginInvoke(new Action(() =>
                {
                    while (listView1.Items.Count <= nv_recv)
                    {
                        listView1.Items.Add(listView1.Items.Count.ToString("X2"));
                        listView1.Items[listView1.Items.Count - 1].SubItems.Add("");
                        listView1.Items[listView1.Items.Count - 1].SubItems.Add("");
                        listView1.Items[listView1.Items.Count - 1].SubItems.Add("");
                        listView1.Items[listView1.Items.Count - 1].SubItems.Add("");
                    }
                    listView1.Items[nv_recv].SubItems[1].Text = name;
                    listView1.Items[nv_recv].SubItems[2].Text = desc;
                    listView1.Items[nv_recv].SubItems[3].Text = hex;
                    listView1.Items[nv_recv].SubItems[4].Text = value;
                    listView1.Items[nv_recv].Tag = nv_recv;
                }));
            }
            catch(Exception ex)
            {

            }
        }

        private void Button1_Click(object sender, EventArgs e)
        {
            if (!int.TryParse(txtAddress.Text, out Address))
            {
                textBox1.Text = "Failed to parse address";
                return;
            }
            textBox1.Text = "Scanning address " + Address;

            ScanThread = new Thread(ThreadMain);
            ScanThread.Start();
        }
    }
}
